# SPDX-License-Identifier: Apache-2.0
# pylint: disable=R0801

# Standard
from pathlib import Path
from unittest.mock import patch
import json
import os
import platform
import sys
import typing

# Third Party
from click.testing import CliRunner
import pytest

# First Party
from instructlab import lab
from instructlab.configuration import DEFAULTS
from instructlab.train import linux_train

INPUT_DIR = "test_generated"
TRAINING_RESULTS_DIR = "training_results"
CHECKPOINT_DIR_NAME = "checkpoint-1"
MERGED_MODEL_DIR_NAME = "merged_model"
FINAL_RESULTS_DIR_NAME = "final"
CHECKPOINT_DIR = TRAINING_RESULTS_DIR + "/" + CHECKPOINT_DIR_NAME
MERGED_MODEL_DIR = TRAINING_RESULTS_DIR + "/" + MERGED_MODEL_DIR_NAME
FINAL_RESULTS_DIR = TRAINING_RESULTS_DIR + "/" + FINAL_RESULTS_DIR_NAME
LINUX_GGUF_FILE = FINAL_RESULTS_DIR + "/ggml-model-f16.gguf"
MODEL_DIR = "model"
ENCODING = "UTF-8"


def setup_input_dir(root: typing.Optional[str] = None):
    input_dir = root if root else INPUT_DIR
    os.makedirs(input_dir)
    for f_path in ["train_1.jsonl", "test_1.jsonl"]:
        print(f"Creating {os.path.join(input_dir, f_path)}")
        with open(os.path.join(input_dir, f_path), "w", encoding=ENCODING):
            pass


def setup_linux_dir():
    os.makedirs(CHECKPOINT_DIR)
    for f_path in [
        "added_tokens.json",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer.model",
        "tokenizer_config.json",
    ]:
        with open(os.path.join(CHECKPOINT_DIR, f_path), "w", encoding=ENCODING) as f:
            f.write("{}")
    os.makedirs(MERGED_MODEL_DIR)
    for f_path in ["config.json", "generation_config.json", "1.safetensors"]:
        with open(os.path.join(MERGED_MODEL_DIR, f_path), "w", encoding=ENCODING) as f:
            f.write("{}")


def setup_load(root: typing.Optional[str] = None):
    model_dir = os.path.join(root, MODEL_DIR) if root else MODEL_DIR
    os.makedirs(model_dir)


def is_arm_mac():
    return sys.platform == "darwin" and platform.machine() == "arm64"


def mock_convert_llama_to_gguf(model, pad_vocab):
    with open(LINUX_GGUF_FILE, "w", encoding="utf-8") as fp:
        fp.write(str(model) + str(pad_vocab))
    return LINUX_GGUF_FILE


@pytest.mark.usefixtures("mock_mlx_package")
class TestLabTrain:
    """Test collection for `ilab model train` command."""

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=True)
    @patch("instructlab.mlx_explore.gguf_convert_to_mlx.load")
    @patch("instructlab.train.lora_mlx.make_data.make_data")
    @patch("instructlab.train.lora_mlx.convert.convert_between_mlx_and_pytorch")
    @patch("instructlab.train.lora_mlx.lora.load_and_train")
    def test_train_mac(
        self,
        load_and_train_mock,
        convert_between_mlx_and_pytorch_mock,
        make_data_mock,
        load_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir()
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                INPUT_DIR,
            ],
        )
        assert result.exit_code == 0
        load_mock.assert_not_called()
        load_and_train_mock.assert_called_once()
        assert load_and_train_mock.call_args[1]["model"] is not None
        assert load_and_train_mock.call_args[1]["train"]
        assert load_and_train_mock.call_args[1]["data"] == DEFAULTS.DATASETS_DIR
        assert load_and_train_mock.call_args[1]["adapter_file"] is not None
        assert load_and_train_mock.call_args[1]["iters"] == 100
        assert load_and_train_mock.call_args[1]["save_every"] == 10
        assert load_and_train_mock.call_args[1]["steps_per_eval"] == 10
        assert len(load_and_train_mock.call_args[1]) == 7
        convert_between_mlx_and_pytorch_mock.assert_called_once()
        assert convert_between_mlx_and_pytorch_mock.call_args[1]["hf_path"] is not None
        assert convert_between_mlx_and_pytorch_mock.call_args[1]["mlx_path"] is not None
        assert convert_between_mlx_and_pytorch_mock.call_args[1]["quantize"]
        assert not convert_between_mlx_and_pytorch_mock.call_args[1]["local"]
        assert len(convert_between_mlx_and_pytorch_mock.call_args[1]) == 4
        make_data_mock.assert_called_once()
        assert make_data_mock.call_args[1]["data_dir"] == DEFAULTS.DATASETS_DIR
        assert len(make_data_mock.call_args[1]) == 1
        is_macos_with_m_chip_mock.assert_called_once()

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=True)
    @patch("instructlab.mlx_explore.gguf_convert_to_mlx.load")
    @patch("instructlab.train.lora_mlx.make_data.make_data")
    @patch("instructlab.train.lora_mlx.convert.convert_between_mlx_and_pytorch")
    @patch("instructlab.train.lora_mlx.lora.load_and_train")
    def test_skip_quantize(
        self,
        load_and_train_mock,
        convert_between_mlx_and_pytorch_mock,
        make_data_mock,
        load_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir()
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                INPUT_DIR,
                "--skip-quantize",
            ],
        )
        assert result.exit_code == 0
        load_mock.assert_not_called()
        load_and_train_mock.assert_called_once()
        convert_between_mlx_and_pytorch_mock.assert_called_once()
        assert convert_between_mlx_and_pytorch_mock.call_args[1]["quantize"] is False
        make_data_mock.assert_called_once()
        is_macos_with_m_chip_mock.assert_called_once()

    def test_input_error(self, cli_runner: CliRunner):
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                "invalid",
            ],
        )
        assert result.exception is not None
        assert "Could not read directory: invalid" in result.output
        assert result.exit_code == 1

    def test_invalid_taxonomy(self, cli_runner: CliRunner):
        os.mkdir(INPUT_DIR)  # Leave out the test and train files
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                INPUT_DIR,
            ],
        )
        assert result.exception is not None
        assert (
            f"{INPUT_DIR} does not contain training or test files, did you run `ilab data generate`?"
            in result.output
        )
        assert result.exit_code == 1

    def test_invalid_data_dir(self, cli_runner: CliRunner):
        # The error comes from make_data itself so it's only really useful to test on a mac
        if is_arm_mac():
            os.mkdir(INPUT_DIR)  # Leave out the test and train files
            result = cli_runner.invoke(
                lab.ilab,
                [
                    "--config=DEFAULT",
                    "model",
                    "train",
                    "--pipeline",
                    "simple",
                    "--data-path",
                    "invalid",
                    "--input-dir",
                    INPUT_DIR,
                ],
            )
            assert result.exception is not None
            assert "Could not read from data directory" in result.output
            assert result.exit_code == 1

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=True)
    @patch(
        "instructlab.train.lora_mlx.make_data.make_data",
        side_effect=FileNotFoundError(),
    )
    def test_invalid_data_dir_synthetic(
        self, make_data_mock, is_macos_with_m_chip_mock, cli_runner: CliRunner
    ):
        os.mkdir(INPUT_DIR)  # Leave out the test and train files
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--data-path",
                "invalid",
                "--input-dir",
                INPUT_DIR,
            ],
        )
        make_data_mock.assert_called_once()
        assert result.exception is not None
        assert "Could not read from data directory" in result.output
        assert result.exit_code == 1
        is_macos_with_m_chip_mock.assert_called_once()

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=True)
    @patch("instructlab.mlx_explore.gguf_convert_to_mlx.load")
    @patch("instructlab.train.lora_mlx.make_data.make_data")
    @patch("instructlab.train.lora_mlx.convert.convert_between_mlx_and_pytorch")
    @patch("instructlab.train.lora_mlx.lora.load_and_train")
    def test_skip_preprocessing(
        self,
        load_and_train_mock,
        convert_between_mlx_and_pytorch_mock,
        make_data_mock,
        load_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir()
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                INPUT_DIR,
                "--skip-preprocessing",
            ],
        )
        assert result.exit_code == 0
        load_mock.assert_not_called()
        load_and_train_mock.assert_called_once()
        convert_between_mlx_and_pytorch_mock.assert_called_once()
        make_data_mock.assert_not_called()
        is_macos_with_m_chip_mock.assert_called_once()

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=True)
    @patch("instructlab.mlx_explore.utils.fetch_tokenizer_from_hub")
    @patch("instructlab.mlx_explore.gguf_convert_to_mlx.load")
    @patch("instructlab.train.lora_mlx.make_data.make_data")
    @patch("instructlab.train.lora_mlx.convert.convert_between_mlx_and_pytorch")
    @patch("instructlab.train.lora_mlx.lora.load_and_train")
    def test_load(
        self,
        load_and_train_mock,
        convert_between_mlx_and_pytorch_mock,
        make_data_mock,
        load_mock,
        fetch_tokenizer_from_hub_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir(DEFAULTS.DATASETS_DIR)
        setup_load(DEFAULTS.CHECKPOINTS_DIR)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                DEFAULTS.DATASETS_DIR,
                "--tokenizer-dir",
                "tokenizer",
                "--gguf-model-path",
                "gguf_model",
                "--model-path",
                MODEL_DIR,
            ],
        )
        assert result.exit_code == 0
        load_mock.assert_called_once()
        load_and_train_mock.assert_called_once()
        convert_between_mlx_and_pytorch_mock.assert_not_called()
        make_data_mock.assert_called_once()
        fetch_tokenizer_from_hub_mock.assert_called_once()
        assert fetch_tokenizer_from_hub_mock.call_args[0][0] == "tokenizer"
        assert fetch_tokenizer_from_hub_mock.call_args[0][1] == "tokenizer"
        assert len(fetch_tokenizer_from_hub_mock.call_args[0]) == 2
        is_macos_with_m_chip_mock.assert_called_once()

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=True)
    @patch("instructlab.mlx_explore.utils.fetch_tokenizer_from_hub")
    @patch("instructlab.mlx_explore.gguf_convert_to_mlx.load")
    @patch("instructlab.train.lora_mlx.make_data.make_data")
    @patch("instructlab.train.lora_mlx.convert.convert_between_mlx_and_pytorch")
    @patch("instructlab.train.lora_mlx.lora.load_and_train")
    def test_load_local(
        self,
        load_and_train_mock,
        convert_between_mlx_and_pytorch_mock,
        make_data_mock,
        load_mock,
        fetch_tokenizer_from_hub_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir(DEFAULTS.DATASETS_DIR)
        setup_load(DEFAULTS.CHECKPOINTS_DIR)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                DEFAULTS.DATASETS_DIR,
                "--tokenizer-dir",
                "tokenizer",
                "--gguf-model-path",
                "gguf_model",
                "--model-path",
                MODEL_DIR,
                "--local",
            ],
        )
        print(result.output)
        assert result.exit_code == 0
        load_mock.assert_called_once()
        load_and_train_mock.assert_called_once()
        convert_between_mlx_and_pytorch_mock.assert_not_called()
        make_data_mock.assert_called_once()
        fetch_tokenizer_from_hub_mock.assert_not_called()
        is_macos_with_m_chip_mock.assert_called_once()

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=False)
    @patch.object(linux_train, "linux_train", return_value=Path("training_results"))
    @patch(
        "instructlab.llamacpp.llamacpp_convert_to_gguf.convert_llama_to_gguf",
        side_effect=mock_convert_llama_to_gguf,
    )
    def test_train_linux(
        self,
        convert_llama_to_gguf_mock,
        linux_train_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir()
        setup_linux_dir()
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                INPUT_DIR,
            ],
        )
        assert result.exit_code == 0
        convert_llama_to_gguf_mock.assert_called_once()
        assert convert_llama_to_gguf_mock.call_args[1]["model"] == Path(
            "training_results/final"
        )
        assert convert_llama_to_gguf_mock.call_args[1]["pad_vocab"] is True
        assert len(convert_llama_to_gguf_mock.call_args[1]) == 2
        linux_train_mock.assert_called_once()
        print(linux_train_mock.call_args[1])
        assert linux_train_mock.call_args[1]["train_file"] == Path(
            os.path.join(DEFAULTS.DATASETS_DIR, "train_gen.jsonl")
        )
        assert linux_train_mock.call_args[1]["test_file"] == Path(
            os.path.join(DEFAULTS.DATASETS_DIR, "test_gen.jsonl")
        )
        assert linux_train_mock.call_args[1]["num_epochs"] == 10
        assert linux_train_mock.call_args[1]["train_device"] is not None
        assert not linux_train_mock.call_args[1]["four_bit_quant"]
        assert len(linux_train_mock.call_args[1]) == 6
        is_macos_with_m_chip_mock.assert_called_once()
        assert not os.path.isfile(LINUX_GGUF_FILE)

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=False)
    @patch.object(linux_train, "linux_train", return_value=Path("training_results"))
    @patch(
        "instructlab.llamacpp.llamacpp_convert_to_gguf.convert_llama_to_gguf",
        side_effect=mock_convert_llama_to_gguf,
    )
    def test_double_train_linux(
        self,
        convert_llama_to_gguf_mock,
        linux_train_mock,
        is_macos_with_m_chip_mock,
        tmp_path_home,
    ):
        cli_runner = CliRunner()
        with cli_runner.isolated_filesystem(temp_dir=tmp_path_home):
            # re-initialize the defaults
            setup_input_dir(DEFAULTS.DATASETS_DIR)
            setup_input_dir()
            setup_linux_dir()

            # create the files so they already exist
            test_legacy_message = json.dumps(
                {"user": "hi", "system": "hi", "assistant": "hi"}
            )
            test_files_to_create = ["train_gen.jsonl", "test_gen.jsonl"]
            for f in test_files_to_create:
                with open(
                    os.path.join(tmp_path_home, DEFAULTS.DATASETS_DIR, f),
                    "w",
                    encoding=ENCODING,
                ) as outfile:
                    outfile.write(f"{test_legacy_message}\n")

            assert "train_gen.jsonl" in os.listdir(DEFAULTS.DATASETS_DIR)
            assert "test_gen.jsonl" in os.listdir(DEFAULTS.DATASETS_DIR)
            result = cli_runner.invoke(
                lab.ilab,
                [
                    "--config=DEFAULT",
                    "model",
                    "train",
                    "--pipeline",
                    "simple",
                    "--input-dir",
                    DEFAULTS.DATASETS_DIR,
                    "--data-path",
                    DEFAULTS.DATASETS_DIR,
                ],
            )
            assert result.exception is None
            assert result.exit_code == 0
            convert_llama_to_gguf_mock.assert_called_once()
            assert convert_llama_to_gguf_mock.call_args[1]["model"] == Path(
                "training_results/final"
            )
            assert convert_llama_to_gguf_mock.call_args[1]["pad_vocab"] is True
            assert len(convert_llama_to_gguf_mock.call_args[1]) == 2
            linux_train_mock.assert_called_once()
            print(linux_train_mock.call_args[1])
            assert linux_train_mock.call_args[1]["train_file"] == Path(
                os.path.join(DEFAULTS.DATASETS_DIR, "train_gen.jsonl")
            )
            assert linux_train_mock.call_args[1]["test_file"] == Path(
                os.path.join(DEFAULTS.DATASETS_DIR, "test_gen.jsonl")
            )
            assert linux_train_mock.call_args[1]["num_epochs"] == 10
            assert linux_train_mock.call_args[1]["train_device"] is not None
            assert not linux_train_mock.call_args[1]["four_bit_quant"]
            assert len(linux_train_mock.call_args[1]) == 6
            is_macos_with_m_chip_mock.assert_called_once()
            assert not os.path.isfile(LINUX_GGUF_FILE)

            assert "train_gen.jsonl" in os.listdir(DEFAULTS.DATASETS_DIR)
            assert "test_gen.jsonl" in os.listdir(DEFAULTS.DATASETS_DIR)
            # run this test a second time to ensure files are getting selected correctly
            result = cli_runner.invoke(
                lab.ilab,
                [
                    "--config=DEFAULT",
                    "model",
                    "train",
                    "--pipeline",
                    "simple",
                    "--input-dir",
                    DEFAULTS.DATASETS_DIR,
                    "--data-path",
                    DEFAULTS.DATASETS_DIR,
                ],
            )
            # assert result.exit_code == 0
            assert result.exception is None
            assert convert_llama_to_gguf_mock.call_args[1]["model"] == Path(
                "training_results/final"
            )
            assert convert_llama_to_gguf_mock.call_args[1]["pad_vocab"] is True
            assert len(convert_llama_to_gguf_mock.call_args[1]) == 2
            print(linux_train_mock.call_args[1])
            assert linux_train_mock.call_args[1]["train_file"] == Path(
                os.path.join(DEFAULTS.DATASETS_DIR, "train_gen.jsonl")
            )
            assert linux_train_mock.call_args[1]["test_file"] == Path(
                os.path.join(DEFAULTS.DATASETS_DIR, "test_gen.jsonl")
            )
            assert linux_train_mock.call_args[1]["num_epochs"] == 10
            assert linux_train_mock.call_args[1]["train_device"] is not None
            assert not linux_train_mock.call_args[1]["four_bit_quant"]
            assert len(linux_train_mock.call_args[1]) == 6
            assert not os.path.isfile(LINUX_GGUF_FILE)

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=False)
    @patch(
        "instructlab.train.linux_train.linux_train",
        return_value=Path("training_results"),
    )
    @patch(
        "instructlab.llamacpp.llamacpp_convert_to_gguf.convert_llama_to_gguf",
        side_effect=mock_convert_llama_to_gguf,
    )
    def test_num_epochs(
        self,
        convert_llama_to_gguf_mock,
        linux_train_mock,
        is_macos_with_m_chip_mock,
        cli_runner: CliRunner,
    ):
        setup_input_dir()
        setup_linux_dir()
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--pipeline",
                "simple",
                "--input-dir",
                INPUT_DIR,
                "--num-epochs",
                "2",
            ],
        )
        assert result.exit_code == 0
        convert_llama_to_gguf_mock.assert_called_once()
        linux_train_mock.assert_called_once()
        assert linux_train_mock.call_args[1]["num_epochs"] == 2
        is_macos_with_m_chip_mock.assert_called_once()
        assert not os.path.isfile(LINUX_GGUF_FILE)

        # Test with invalid num_epochs
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "train",
                "--input-dir",
                INPUT_DIR,
                "--num-epochs",
                "two",
            ],
        )
        assert result.exception is not None
        assert result.exit_code == 2
        assert "'two' is not a valid integer" in result.output



# SPDX-License-Identifier: Apache-2.0

# Standard
import datetime
import json
import logging
import os
import pathlib
import sys
import time
import traceback

# Third Party
from openai import OpenAI
from prompt_toolkit import PromptSession
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.history import FileHistory
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text
import click
import httpx
import openai

# First Party
from instructlab import clickext
from instructlab import client as ilabclient
from instructlab import configuration as cfg
from instructlab import log
from instructlab.utils import HttpClientParams

# Local
from ..utils import get_sysprompt, http_client

logger = logging.getLogger(__name__)

HELP_MD = """
Help / TL;DR
- `/q`: **q**uit
- `/h`: show **h**elp
- `/a assistant`: **a**mend **a**ssistant (i.e., model)
- `/c context`: **c**hange **c**ontext (available contexts: default, cli_helper)
- `/lc`: **l**ist **c**ontexts
- `/m`: toggle **m**ultiline (for the next session only)
- `/M`: toggle **m**ultiline
- `/n`: **n**ew session
- `/N`: **n**ew session (ignoring loaded)
- `/d <int>`: **d**isplay previous response based on input, if passed 1 then previous, if 2 then second last response and so on.
- `/p <int>`: previous response in **p**lain text based on input, if passed 1 then previous, if 2 then second last response and so on.
- `/md <int>`: previous response in **M**ark**d**own based on input, if passed 1 then previous, if 2 then second last response and so on.
- `/s filepath`: **s**ave current session to `filepath`
- `/l filepath`: **l**oad `filepath` and start a new session
- `/L filepath`: **l**oad `filepath` (permanently) and start a new session

Press Alt (or Meta) and Enter or Esc Enter to end multiline input.
"""

CONTEXTS = {
    "default": get_sysprompt(),
    "cli_helper": "You are an expert for command line interface and know all common commands. Answer the command to execute as it without any explanation.",
}

PROMPT_HISTORY_FILEPATH = os.path.expanduser("~/.local/chat-cli.history")

PROMPT_PREFIX = ">>> "

DEFAULT_ENDPOINT = cfg._serve().api_base()


def is_openai_server_and_serving_model(
    endpoint: str, api_key: str, http_params: HttpClientParams
) -> bool:
    """
    Given an endpoint, returns whether or not the server is OpenAI-compatible
    and is actively serving at least one model.
    """
    try:
        models = ilabclient.list_models(
            endpoint, api_key=api_key, http_client=http_client(http_params)
        )
        return len(models.data) > 0
    except ilabclient.ClientException:
        return False


@click.command()
@click.argument(
    "question",
    nargs=-1,
    type=click.UNPROCESSED,
)
@click.option(
    "-m",
    "--model",
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "-c",
    "--context",
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "-s",
    "--session",
    type=click.File("r"),
    cls=clickext.ConfigOption,
)
@click.option(
    "-qq",
    "--quick-question",
    is_flag=True,
    help="Exit after answering question.",
)
@click.option(
    "-gm",
    "--greedy-mode",
    is_flag=True,
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--max-tokens",
    type=click.INT,
    cls=clickext.ConfigOption,
)
@click.option(
    "--endpoint-url",
    type=click.STRING,
    help="Custom URL endpoint for OpenAI-compatible API. Defaults to the `ilab model serve` endpoint.",
)
@click.option(
    "--api-key",
    type=click.STRING,
    default=cfg.DEFAULTS.API_KEY,  # Note: do not expose default API key
    help="API key for API endpoint. [default: config.DEFAULT_API_KEY]",
)
@click.option(
    "--tls-insecure",
    is_flag=True,
    help="Disable TLS verification.",
)
@click.option(
    "--tls-client-cert",
    type=click.Path(),
    default="",
    show_default=True,
    help="Path to the TLS client certificate to use.",
)
@click.option(
    "--tls-client-key",
    type=click.Path(),
    default="",
    show_default=True,
    help="Path to the TLS client key to use.",
)
@click.option(
    "--tls-client-passwd",
    type=click.STRING,
    default="",
    help="TLS client certificate password.",
)
@click.option(
    "--model-family",
    help="Force model family to use when picking a chat template",
)
@click.option(
    "--serving-log-file",
    type=click.Path(path_type=pathlib.Path),
    required=False,
    help="Log file path to write server logs to.",
)
@click.pass_context
@clickext.display_params
def chat(
    ctx,
    question,
    model,
    context,
    session,
    quick_question,
    greedy_mode,
    max_tokens,
    endpoint_url,
    api_key,
    tls_insecure,
    tls_client_cert,
    tls_client_key,
    tls_client_passwd,
    model_family,
    serving_log_file,
):
    """Runs a chat using the modified model"""
    # pylint: disable=import-outside-toplevel
    # First Party
    from instructlab.model.backends.common import is_temp_server_running

    users_endpoint_url = cfg.get_api_base(ctx.obj.config.serve.host_port)

    # we prefer the given endpoint when one is provided, else we check if the user
    # is actively serving something before falling back to serving our own model
    backend_instance = None
    if endpoint_url:
        api_base = endpoint_url
    elif is_openai_server_and_serving_model(
        users_endpoint_url,
        api_key,
        http_params={
            "tls_client_cert": tls_client_cert,
            "tls_client_key": tls_client_key,
            "tls_client_passwd": tls_client_passwd,
            "tls_insecure": tls_insecure,
        },
    ):
        api_base = users_endpoint_url
        if serving_log_file:
            logger.warning(
                "Setting serving log file (--serving-log-file) is not supported when the server is already running"
            )
    else:
        # First Party
        from instructlab.model.backends import backends

        # If a log file is specified, write logs to the file
        root_logger = logging.getLogger()
        if serving_log_file:
            log.add_file_handler_to_logger(root_logger, serving_log_file)

        ctx.obj.config.serve.llama_cpp.llm_family = model_family
        backend_instance = backends.select_backend(
            ctx.obj.config.serve,
            model_path=model,
            log_file=serving_log_file,
        )
        try:
            # Run the llama server
            api_base = backend_instance.run_detached(http_client(ctx.params))
        except Exception as exc:
            click.secho(f"Failed to start server: {exc}", fg="red")
            raise click.exceptions.Exit(1)

    # if only the chat is running (`ilab model chat`) and the temp server is not, the chat interacts
    # in server mode (`ilab model serve` is running somewhere, or we are talking to another
    # OpenAI compatible endpoint).
    if not is_temp_server_running():
        # Try to get the model name right if we know we're talking to a local `ilab model serve`.
        #
        # If the model from the CLI and the one in the config are the same, use the one from the
        # server if they are different else let's use what the user provided
        #
        # 'model' will always get a value and never be None so it's hard to distinguish whether
        # the value came from the user input or the default value.
        # We can only assume that if the value is the same as the default value and the value
        # from the config is the same as the default value, then the user didn't provide a value
        # we then compare it with the value from the server to see if it's different
        if (
            # We need to get the base name of the model because the model path is a full path and
            # the once from the config is just the model name
            os.path.basename(model) == cfg.DEFAULTS.GGUF_MODEL_NAME
            and os.path.basename(ctx.obj.config.chat.model)
            == cfg.DEFAULTS.GGUF_MODEL_NAME
            and api_base == ctx.obj.config.serve.api_base()
        ):
            logger.debug(
                "No model was provided by the user as a CLI argument or in the config, will use the model from the server"
            )
            try:
                models = ilabclient.list_models(
                    api_base=api_base,
                    http_client=http_client(ctx.params),
                )

                # Currently, we only present a single model so we can safely assume that the first model
                server_model = models.data[0].id if models is not None else None

                # override 'model' with the first returned model if not provided so that the chat print
                # the model used by the server
                model = (
                    server_model
                    if server_model is not None
                    and server_model != ctx.obj.config.chat.model
                    else model
                )
                logger.debug(f"Using model from server {model}")
            except ilabclient.ClientException as exc:
                click.secho(
                    f"Failed to list models from {api_base}. Please check the API key and endpoint.",
                    fg="red",
                )
                # Right now is_temp_server() does not check if a subprocessed vllm is up
                # shut it down just in case an exception is raised in the try
                # TODO: revise is_temp_server to check if a vllm server is running
                if backend_instance is not None:
                    backend_instance.shutdown()
                raise click.exceptions.Exit(1) from exc

    try:
        chat_cli(
            ctx,
            api_base=api_base,
            config=ctx.obj.config.chat,
            question=question,
            model=model,
            context=context,
            session=session,
            qq=quick_question,
            greedy_mode=greedy_mode,
            max_tokens=max_tokens,
        )
    except ChatException as exc:
        click.secho(f"Executing chat failed with: {exc}", fg="red")
        raise click.exceptions.Exit(1)
    finally:
        if backend_instance is not None:
            backend_instance.shutdown()


class ChatException(Exception):
    """An exception raised during chat step."""


class ChatQuitException(Exception):
    """A quit command was executed during chat."""


# TODO Autosave chat history
class ConsoleChatBot:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        model,
        client,
        vi_mode=False,
        prompt=True,
        vertical_overflow="ellipsis",
        loaded=None,
        log_file=None,
        greedy_mode=False,
        max_tokens=None,
    ):
        self.client = client
        self.model = model
        self.vi_mode = vi_mode
        self.vertical_overflow = vertical_overflow
        self.loaded = loaded
        self.log_file = log_file
        self.greedy_mode = greedy_mode
        self.max_tokens = max_tokens

        self.console = Console()

        self.input = None
        if prompt:
            os.makedirs(os.path.dirname(PROMPT_HISTORY_FILEPATH), exist_ok=True)
            self.input = PromptSession(history=FileHistory(PROMPT_HISTORY_FILEPATH))
        self.multiline = False
        self.multiline_mode = 0

        self.info = {}
        self._reset_session()

    def _reset_session(self, hard=False):
        if hard:
            self.loaded = {}
        self.info["messages"] = (
            []
            if hard or ("messages" not in self.loaded)
            else [*self.loaded["messages"]]
        )

    def _sys_print(self, *args, **kwargs):
        self.console.print(Panel(*args, title="system", **kwargs))

    def log_message(self, msg):
        if self.log_file:
            with open(self.log_file, "a", encoding="utf-8") as fp:
                fp.write(msg)

    def greet(self, help=False, new=False, session_name="new session"):  # pylint: disable=redefined-builtin
        side_info_str = (" (type `/h` for help)" if help else "") + (
            f" ({session_name})" if new else ""
        )
        self._sys_print(
            Markdown(
                f"Welcome to InstructLab Chat w/ **{self.model_name.upper()}**"
                + side_info_str
            )
        )

    @property
    def model_name(self):
        return os.path.basename(os.path.normpath(self.model))

    @property
    def _right_prompt(self):
        return FormattedText(
            [
                (
                    "#3f7cac bold",
                    f"[{'M' if self.multiline else 'S'}]",
                ),  # info blue for multiple
                *(
                    [("bold", f"[{self.loaded['name']}]")]
                    if "name" in self.loaded
                    else []
                ),  # loaded context/session file
                # TODO: Fix openai package to fix the openai error.
                # *([] if openai.proxy is None else [('#d08770 bold', "[proxied]")]), # indicate prox
            ]
        )

    def _handle_quit(self, _):
        raise ChatQuitException

    def _handle_help(self, _):
        self._sys_print(Markdown(HELP_MD))
        raise KeyboardInterrupt

    def _handle_multiline(self, content):
        temp = content == "/m"  # soft multiline only for next prompt
        self.multiline = not self.multiline
        self.multiline_mode = 1 if not temp else 2
        raise KeyboardInterrupt

    def _handle_amend(self, content):
        cs = content.split()
        if len(cs) < 2:
            self._sys_print(
                Markdown(
                    "**WARNING**: The second argument `assistant` is missing in the `/a assistant` command."
                )
            )
            raise KeyboardInterrupt
        self.model = cs[1]
        self._reset_session()
        self.greet(new=True)
        raise KeyboardInterrupt

    def _handle_context(self, content):
        if CONTEXTS is None:
            self._sys_print(
                Markdown("**WARNING**: No contexts loaded from the config file.")
            )
            raise KeyboardInterrupt
        cs = content.split()
        if len(cs) < 2:
            self._sys_print(
                Markdown(
                    "**WARNING**: The second argument `context` is missing in the `/c context` command."
                )
            )
            raise KeyboardInterrupt
        context = cs[1]
        if context not in CONTEXTS:
            available_contexts = ", ".join(CONTEXTS.keys())
            self._sys_print(
                Markdown(
                    f"**WARNING**: Context `{context}` not found. "
                    f"Available contexts: `{available_contexts}`"
                )
            )
            raise KeyboardInterrupt
        self.loaded["name"] = context
        self.loaded["messages"] = [{"role": "system", "content": CONTEXTS[context]}]
        self._reset_session()
        self.greet(new=True)
        raise KeyboardInterrupt

    def _handle_new_session(self, content):
        hard = content == "/N"  # hard new ignores loaded context/session
        self._reset_session(hard=hard)
        self.greet(new=True)
        raise KeyboardInterrupt

    def __handle_replay(self, content, display_wrapper=lambda x: x):
        # if the history is empty, then return
        if (
            len(self.info["messages"]) == 1
            and self.info["messages"][0]["role"] == "system"
        ):
            raise KeyboardInterrupt
        cs = content.split()
        try:
            i = 1 if len(cs) == 1 else int(cs[1]) * 2 - 1
            if abs(i) >= len(self.info["messages"]):
                raise IndexError
        except (IndexError, ValueError) as exc:
            self.console.print(
                display_wrapper("Invalid index: " + content), style="bold red"
            )
            raise KeyboardInterrupt from exc
        if len(self.info["messages"]) > abs(i):
            self.console.print(display_wrapper(self.info["messages"][-i]["content"]))
        raise KeyboardInterrupt

    def _handle_display(self, content):
        return self.__handle_replay(content, display_wrapper=lambda x: Panel(x))  # pylint: disable=unnecessary-lambda

    def _load_session_history(self, content=None):
        data = self.info["messages"]
        if content is not None:
            data = content["messages"]
        for m in data:
            if m["role"] == "user":
                self.console.print(
                    "\n" + PROMPT_PREFIX + m["content"], style="dim grey0"
                )
            else:
                self.console.print(Panel(m["content"]), style="dim grey0")

    def _handle_plain(self, content):
        return self.__handle_replay(content)

    def _handle_markdown(self, content):
        return self.__handle_replay(
            content,
            display_wrapper=(
                lambda x: Panel(
                    Markdown(x), subtitle_align="right", subtitle="rendered as Markdown"
                )
            ),
        )

    def _handle_save_session(self, content):
        cs = content.split()
        if len(cs) < 2:
            self._sys_print(
                Markdown(
                    "**WARNING**: The second argument `filepath` is missing in the `/s filepath` command."
                )
            )
            raise KeyboardInterrupt
        filepath = cs[1]
        with open(filepath, "w", encoding="utf-8") as outfile:
            json.dump(self.info["messages"], outfile, indent=4)
        raise KeyboardInterrupt

    def _handle_load_session(self, content):
        cs = content.split()
        if len(cs) < 2:
            self._sys_print(
                Markdown(
                    "**WARNING**: The second argument `filepath` is missing in the `/l filepath` or `/L filepath` command."
                )
            )
            raise KeyboardInterrupt
        filepath = cs[1]
        if not os.path.exists(filepath):
            self._sys_print(
                Markdown(
                    f"**WARNING**: File `{filepath}` specified in the `/l filepath` or `/L filepath` command does not exist."
                )
            )
            raise KeyboardInterrupt
        with open(filepath, "r", encoding="utf-8") as session:
            messages = json.loads(session.read())
        if content[:2] == "/L":
            self.loaded["name"] = filepath
            self.loaded["messages"] = messages
            self._reset_session()
            self.greet(new=True)
        else:
            self._reset_session()
            self.info["messages"] = [*messages]
            self.greet(new=True, session_name=filepath)

        # now load session's history
        self._load_session_history()
        raise KeyboardInterrupt

    def _handle_empty(self):
        raise KeyboardInterrupt

    def _update_conversation(self, content, role):
        assert role in ("user", "assistant")
        message = {"role": role, "content": content}
        self.info["messages"].append(message)

    def _handle_list_contexts(self, _):
        context_list = "\n\n".join(
            [f"**{key}**:\n{value}" for key, value in CONTEXTS.items()]
        )
        self._sys_print(Markdown(f"**Available contexts:**\n\n{context_list}"))
        raise KeyboardInterrupt

    def start_prompt(
        self,
        logger,  # pylint: disable=redefined-outer-name
        content=None,
        box=True,
    ):
        handlers = {
            "/q": self._handle_quit,
            "quit": self._handle_quit,
            "exit": self._handle_quit,
            "/h": self._handle_help,
            "/a": self._handle_amend,
            "/c": self._handle_context,
            "/m": self._handle_multiline,
            "/n": self._handle_new_session,
            "/d": self._handle_display,
            "/p": self._handle_plain,
            "/md": self._handle_markdown,
            "/s": self._handle_save_session,
            "/l": self._handle_load_session,
            "/lc": self._handle_list_contexts,
        }

        if content is None:
            content = self.input.prompt(
                PROMPT_PREFIX,
                rprompt=self._right_prompt,
                vi_mode=True,
                multiline=self.multiline,
            )

        # Handle empty
        if content.strip() == "":
            raise KeyboardInterrupt

        # Handle commands
        handler = handlers.get(content.split()[0].lower(), None)
        if handler is not None:
            handler(content)

        self.log_message(PROMPT_PREFIX + content + "\n\n")

        # Update message history and token counters
        self._update_conversation(content, "user")

        # Deal with temp multiline
        if self.multiline_mode == 2:
            self.multiline_mode = 0
            self.multiline = not self.multiline

        # Optional parameters
        create_params = {}
        if self.greedy_mode:
            # https://platform.openai.com/docs/api-reference/chat/create#chat-create-temperature
            create_params["temperature"] = 0

        if self.max_tokens:
            create_params["max_tokens"] = self.max_tokens

        # Get and parse response
        try:
            while True:
                # Loop to catch situations where we need to retry, such as context length exceeded
                try:
                    response = self.client.chat.completions.create(
                        model=self.model,
                        messages=self.info["messages"],
                        stream=True,
                        **create_params,
                    )
                except openai.BadRequestError as e:
                    logger.debug(f"BadRequestError: {e}")
                    if e.code == "context_length_exceeded":
                        if len(self.info["messages"]) > 1:
                            # Trim the oldest entry in our message history
                            logger.debug(
                                "Trimming message history to attempt to fit context length"
                            )
                            self.info["messages"] = self.info["messages"][1:]
                            continue
                        # We only have a single message and it's still to big.
                        self.console.print(
                            "Message too large for context size.", style="bold red"
                        )
                        self.info["messages"].pop()
                        raise KeyboardInterrupt from e
                except openai.InternalServerError as e:
                    logger.debug(f"InternalServerError: {e}")
                    self.info["messages"].clear()
                    raise KeyboardInterrupt from e
                assert (
                    next(response).choices[0].delta.role == "assistant"
                ), 'first response should be {"role": "assistant"}'
                break
        except openai.AuthenticationError as e:
            self.console.print(
                "Invalid API Key. Please set it in your config file.", style="bold red"
            )
            raise ChatException("API Key Error") from e
        except openai.RateLimitError as e:
            self.console.print(
                "Rate limit or maximum monthly limit exceeded", style="bold red"
            )
            self.info["messages"].pop()
            raise ChatException("Rate limit exceeded") from e
        except openai.APIConnectionError as e:
            logger.debug("Connection error, try again...", exc_info=True)
            self.console.print("Connection error, try again...", style="red bold")
            self.info["messages"].pop()
            raise KeyboardInterrupt from e
        except KeyboardInterrupt as e:
            raise e
        except httpx.RemoteProtocolError as e:
            logger.debug("Connection to the server was closed", exc_info=True)
            self.console.print("Connection to the server was closed", style="bold red")
            self.info["messages"].pop()
            raise ChatException("Connection to the server was closed") from e
        except Exception as e:
            if logger.getEffectiveLevel() == logging.DEBUG:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                message = f"Unknown error: {''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))}"
            else:
                message = f"Unknown error: {sys.exc_info()[0]}"

            self.console.print("Unknown error", style="bold red")
            raise ChatException(message) from e

        response_content = Text()
        panel = (
            Panel(
                response_content,
                title=self.model_name,
                subtitle_align="right",
            )
            if box
            else response_content
        )
        subtitle = None
        with Live(
            panel,
            console=self.console,
            refresh_per_second=5,
            vertical_overflow=self.vertical_overflow,
        ):
            start_time = time.time()
            for chunk in response:
                chunk_message = chunk.choices[0].delta
                if chunk_message.content:
                    response_content.append(chunk_message.content)

                if box:
                    panel.subtitle = f"elapsed {time.time() - start_time:.3f} seconds"
            subtitle = f"elapsed {time.time() - start_time:.3f} seconds"

        # Update chat logs
        if subtitle is not None:
            self.log_message("- " + subtitle + " -\n")
        self.log_message(response_content.plain + "\n\n")
        # Update message history and token counters
        self._update_conversation(response_content.plain, "assistant")


def chat_cli(
    ctx,
    api_base,
    config,
    question,
    model,
    context,
    session,
    qq,
    greedy_mode,
    max_tokens,
):
    """Starts a CLI-based chat with the server"""
    client = OpenAI(
        base_url=api_base,
        api_key=ctx.params["api_key"],
        timeout=cfg.DEFAULTS.CONNECTION_TIMEOUT,
        http_client=http_client(ctx.params),
    )
    # ensure the model specified exists on the server. with backends like vllm, this is crucial.
    try:
        model_list = client.models.list().data
    except openai.OpenAIError as exc:
        raise ChatException(f"Is the server running? {exc}") from exc
    model_ids = []
    for m in model_list:
        model_ids.append(m.id)
    if not any(model == m for m in model_ids):
        if model == cfg.DEFAULTS.MODEL_NAME_OLD:
            logger.info(
                f"Model {model} is not a full path. Try running ilab config init or edit your config to have the full model path for serving, chatting, and generation."
            )
        raise ChatException(
            f"Model {model} is not served by the server. These are the served models: {model_ids}"
        )

    # Load context/session
    loaded = {}

    # Context config file
    # global CONTEXTS
    # CONTEXTS = config["contexts"]
    if context not in CONTEXTS:
        logger.info(f"Context {context} not found in the config file. Using default.")
        context = "default"
    loaded["name"] = context
    loaded["messages"] = [{"role": "system", "content": CONTEXTS[context]}]

    # Session from CLI
    if session is not None:
        loaded["name"] = os.path.basename(session.name).strip(".json")
        try:
            loaded["messages"] = json.loads(session.read())
        except json.JSONDecodeError as exc:
            raise ChatException(
                f"Session file {session.name} is not a valid JSON file."
            ) from exc

    log_file = None
    if config.logs_dir:
        date_suffix = (
            datetime.datetime.now().replace(microsecond=0).isoformat().replace(":", "_")
        )
        os.makedirs(config.logs_dir, exist_ok=True)
        log_file = f"{config.logs_dir}/chat_{date_suffix}.log"

    # Initialize chat bot
    ccb = ConsoleChatBot(
        config.model if model is None else model,
        client=client,
        vi_mode=config.vi_mode,
        log_file=log_file,
        prompt=not qq,
        vertical_overflow=("visible" if config.visible_overflow else "ellipsis"),
        loaded=loaded,
        greedy_mode=(
            greedy_mode if greedy_mode else config.greedy_mode
        ),  # The CLI flag can only be used to enable
        max_tokens=(max_tokens if max_tokens else config.max_tokens),
    )

    if not qq and session is None:
        # Greet
        ccb.greet(help=True)

    # Use the input question to start with
    if len(question) > 0:
        question = " ".join(question)
        if not qq:
            print(f"{PROMPT_PREFIX}{question}")
        try:
            ccb.start_prompt(logger, content=question, box=not qq)
        except ChatException as exc:
            raise ChatException(f"API issue found while executing chat: {exc}") from exc
        except (ChatQuitException, KeyboardInterrupt, EOFError):
            return

    if qq:
        return

    # load the history
    if session is not None:
        ccb._load_session_history(loaded)

    # Start chatting
    while True:
        try:
            ccb.start_prompt(logger)
        except KeyboardInterrupt:
            continue
        except ChatException as exc:
            raise ChatException(f"API issue found while executing chat: {exc}") from exc
        except httpx.RemoteProtocolError as exc:
            raise ChatException("Connection to the server was closed") from exc
        except (ChatQuitException, EOFError):
            return



# SPDX-License-Identifier: Apache-2.0
# Third Party
import click

# First Party
from instructlab import clickext


@click.command()
@click.pass_context
@clickext.display_params
def edit(
    ctx,
):
    """Launch $EDITOR to edit the configuration file."""
    click.edit(filename=ctx.obj.config_file)



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
import logging
import os
import typing

# Third Party
# https://huggingface.co/docs/datasets/en/package_reference/loading_methods#datasets.load_dataset
from datasets import load_dataset
from peft import LoraConfig
from tqdm import tqdm

# https://huggingface.co/docs/transformers/model_doc/auto#transformers.AutoConfig
# https://huggingface.co/docs/transformers/model_doc/auto#transformers.AutoModelForCausalLM
# https://huggingface.co/docs/transformers/main_classes/quantization#transformers.BitsAndBytesConfig
# https://huggingface.co/docs/transformers/internal/generation_utils#transformers.StoppingCriteria
# https://huggingface.co/docs/transformers/main_classes/trainer#transformers.TrainingArguments
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    StoppingCriteria,
    StoppingCriteriaList,
    TrainingArguments,
)

# Transformer Reinforcement Learning
# https://huggingface.co/docs/trl/index
from trl import DataCollatorForCompletionOnlyLM, SFTTrainer
import torch

# First Party
from instructlab.utils import ensure_legacy_dataset

# Local
from ..model.chat import CONTEXTS

logger = logging.getLogger(__name__)
# TODO CPU: Look into using these extensions
# import intel_extension_for_pytorch as ipex

# Habana Labs framework for Intel Gaudi HPUs
# The imports register `hpu` device and `torch.hpu` package.
try:
    # pylint: disable=import-error
    # Third Party
    from habana_frameworks.torch import core as htcore
    from habana_frameworks.torch import hpu

    # Habana implementations of SFT Trainer
    # https://huggingface.co/docs/optimum/habana/index
    from optimum.habana import GaudiConfig, GaudiTrainingArguments
    from optimum.habana.trl import GaudiSFTTrainer

    # silence warning: "Call mark_step function will not have any effect"
    logging.getLogger("habana_frameworks.torch.utils.internal").setLevel(logging.ERROR)
except ImportError:
    htcore = None
    hpu = None
    hpu_backends = None


class StoppingCriteriaSub(StoppingCriteria):
    # pylint: disable=unused-argument
    def __init__(self, stops=(), encounters=1, *, device: torch.device):
        super().__init__()
        self.device = device
        self.stops = [stop.to(device) for stop in stops]

    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs
    ) -> bool:
        for seqs in input_ids:
            seq = seqs[-1].to(self.device)
            for stop in self.stops:
                if stop == seq:
                    return True
        return False


def create_prompt(
    user: str,
    system: str = CONTEXTS["default"],
):
    return f"""\
    <|system|>
    {system}
    <|user|>
    {user}
    <|assistant|>
    """


def formatting_prompts_func(example):
    output_texts = []
    for i in range(len(example["system"])):
        text = f"<|system|>\n{example['system'][i]}\n<|user|>\n{example['user'][i]}\n<|assistant|>\n{example['assistant'][i]}<|endoftext|>"
        output_texts.append(text)
    return output_texts


def report_cuda_device(args_device: torch.device, min_vram: int = 0) -> None:
    """Report CUDA/ROCm device properties"""
    print(f"  NVidia CUDA version: {torch.version.cuda or 'n/a'}")
    print(f"  AMD ROCm HIP version: {torch.version.hip or 'n/a'}")

    def _gib(size: int) -> str:
        return "{:.1f} GiB".format(size / 1024**3)

    for idx in range(torch.cuda.device_count()):
        device = torch.device("cuda", idx)
        name = torch.cuda.get_device_name(device)
        free, total = torch.cuda.mem_get_info(device)
        capmin, capmax = torch.cuda.get_device_capability(device)
        print(
            f"  {device} is '{name}' ({_gib(free)} of {_gib(total)} free, "
            f"capability: {capmin}.{capmax})"
        )

    if args_device.index is None:
        index = torch.cuda.current_device()
    else:
        index = args_device.index

    free = torch.cuda.mem_get_info(index)[0]
    if free < min_vram:
        print(
            f"  WARNING: You have less than {min_vram} GiB of free GPU "
            "memory on '{index}'. Training may fail, use slow shared "
            "host memory, or move some layers to CPU."
        )
        print(
            "  Training does not use the local InstructLab serve. Consider "
            "stopping the server to free up about 5 GiB of GPU memory."
        )


def report_hpu_device(args_device: torch.device) -> None:
    print(f"Device count: {hpu.device_count()}")
    for idx in range(hpu.device_count()):
        device = torch.device(args_device.type, idx)
        name: str = hpu.get_device_name(device)
        cap: str = hpu.get_device_capability(device)
        # property string is surrounded by '()'
        prop: str = hpu.get_device_properties(device)
        print(f"  {device} is '{name}', cap: {cap} {prop}")
    # https://docs.habana.ai/en/latest/PyTorch/Reference/Runtime_Flags.html
    print("PT and Habana Environment variables")
    for key, value in sorted(os.environ.items()):
        if key.startswith(("PT_", "HABANA", "LOG_LEVEL_", "ENABLE_CONSOLE")):
            print(f'  {key}="{value}"')


def linux_train(
    train_file: Path,
    test_file: Path,
    model_name: str,
    num_epochs: int | None,
    train_device: str = "cpu",
    four_bit_quant: bool = False,
    output_dir: Path = Path("training_results"),
) -> Path:
    """Lab Train for Linux!"""

    try:
        device = torch.device(train_device)
    except RuntimeError as e:
        raise RuntimeError(f"Error getting Torch Device {str(e)}") from e
    # Detect CUDA/ROCm device
    if device.type == "cuda":
        if not torch.cuda.is_available():
            raise RuntimeError(
                f"{device.type}: Torch has no CUDA/ROCm support or could not detect a compatible device."
            )
        # map unqualified 'cuda' to current device
        if device.index is None:
            device = torch.device(device.type, torch.cuda.current_device())

    if device.type == "hpu":
        logger.warning(
            "WARNING: HPU support is experimental, unstable, and not "
            "optimized, yet.",
        )

    if four_bit_quant and device.type != "cuda":
        raise RuntimeError("'--4-bit-quant' option requires '--device=cuda'")

    print("LINUX_TRAIN.PY: NUM EPOCHS IS: ", num_epochs)
    print("LINUX_TRAIN.PY: TRAIN FILE IS: ", train_file)
    print("LINUX_TRAIN.PY: TEST FILE IS: ", test_file)

    print(f"LINUX_TRAIN.PY: Using device '{device}'")
    output_dir = Path(output_dir)
    if device.type == "cuda":
        # estimated by watching nvtop / radeontop during training
        min_vram = 11 if four_bit_quant else 17

        # convert from gibibytes to bytes, torch.cuda.mem_get_info() returns bytes
        min_vram = min_vram * 1024**3

        report_cuda_device(device, min_vram)
    elif device.type == "hpu":
        if htcore is None:
            raise RuntimeError("habana_framework package is not installed.")
        if not hpu.is_available():
            raise RuntimeError("habana_framework is unable to detect HPUs.")

        hpu.init()
        report_hpu_device(device)

    print("LINUX_TRAIN.PY: LOADING DATASETS")
    # Get the file name
    train_dataset = load_dataset(
        "json", data_files=os.fspath(train_file), split="train"
    )

    test_dataset = load_dataset("json", data_files=os.fspath(test_file), split="train")
    try:
        train_dataset = ensure_legacy_dataset(train_dataset)
        test_dataset = ensure_legacy_dataset(test_dataset)
    except ValueError as e:
        raise ValueError(f"Failed to parse dataset: {e}") from e

    train_dataset.to_pandas().head()

    # https://huggingface.co/docs/transformers/en/model_doc/auto#transformers.AutoTokenizer.from_pretrained
    tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
    tokenizer.pad_token = tokenizer.eos_token

    response_template = "\n<|assistant|>\n"

    response_template_ids = tokenizer.encode(
        response_template, add_special_tokens=False
    )[2:]
    collator = DataCollatorForCompletionOnlyLM(
        response_template_ids, tokenizer=tokenizer
    )

    if four_bit_quant:
        print("LINUX_TRAIN.PY: USING 4-bit quantization with BitsAndBytes")
        use_fp16 = True
        bnb_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=True,
            bnb_4bit_compute_dtype=torch.float16,  # if not set will throw a warning about slow speeds when training
        )
    else:
        print("LINUX_TRAIN.PY: NOT USING 4-bit quantization")
        use_fp16 = False
        bnb_config = None

    # Loading the model
    print("LINUX_TRAIN.PY: LOADING THE BASE MODEL")
    config = AutoConfig.from_pretrained(
        model_name, torchscript=True, trust_remote_code=True
    )

    # https://huggingface.co/docs/transformers/en/model_doc/auto#transformers.AutoModelForCausalLM.from_pretrained
    model = AutoModelForCausalLM.from_pretrained(
        model_name,
        torch_dtype="auto",
        quantization_config=bnb_config,
        config=config,
        trust_remote_code=True,
        low_cpu_mem_usage=True,
    )
    if model.device != device:
        model = model.to(device)
    print(f"LINUX_TRAIN.PY: Model device {model.device}")
    if model.device.type == "cuda":
        print(torch.cuda.memory_summary())

    print("LINUX_TRAIN.PY: SANITY CHECKING THE BASE MODEL")
    stop_words = ["<|endoftext|>", "<|assistant|>"]
    stop_words_ids = [
        tokenizer(stop_word, return_tensors="pt", add_special_tokens=False)[
            "input_ids"
        ].squeeze()
        for stop_word in stop_words
    ]
    stopping_criteria = StoppingCriteriaList(
        [StoppingCriteriaSub(stops=stop_words_ids, device=model.device)]
    )

    def model_generate(user, **kwargs):
        text = create_prompt(user=user)

        input_ids = tokenizer(text, return_tensors="pt").input_ids.to(model.device)
        outputs = model.generate(
            input_ids=input_ids,
            max_new_tokens=256,
            pad_token_id=tokenizer.eos_token_id,
            temperature=0.7,
            top_p=0.9,
            stopping_criteria=stopping_criteria,
            do_sample=True,
            **kwargs,
        )
        return tokenizer.batch_decode([o[:-1] for o in outputs])[0]

    assistant_old_lst = [
        model_generate(d["user"]).split(response_template.strip())[-1].strip()
        for d in tqdm(test_dataset)
    ]
    attention_layers = [
        module for module in model.modules() if "attention" in str(type(module)).lower()
    ]

    print("LINUX_TRAIN.PY: GETTING THE ATTENTION LAYERS")
    # Print information about the attention modules
    for layer in attention_layers:
        for par in list(layer.named_parameters()):
            mod = par[0]
            if isinstance(mod, str):
                mod = mod.split(".")[0]
        break

    print("LINUX_TRAIN.PY: CONFIGURING LoRA")

    lora_alpha = 32
    lora_dropout = 0.1
    lora_r = 4

    target_modules = ["q_proj", "k_proj", "v_proj", "o_proj"]

    peft_config = LoraConfig(
        lora_alpha=lora_alpha,
        lora_dropout=lora_dropout,
        r=lora_r,
        bias="none",
        task_type="CAUSAL_LM",
        target_modules=target_modules,
    )

    tokenizer.padding_side = "right"
    per_device_train_batch_size = 1
    max_seq_length = 300
    generate_kwargs: dict[str, typing.Any]

    if device.type == "hpu":
        # Intel Gaudi trainer
        # https://docs.habana.ai/en/latest/PyTorch/Getting_Started_with_PyTorch_and_Gaudi/Getting_Started_with_PyTorch.html
        # https://huggingface.co/docs/optimum/habana/quickstart
        # https://huggingface.co/docs/optimum/habana/package_reference/gaudi_config
        if per_device_train_batch_size == 1:
            per_device_train_batch_size = 8

        training_arguments = GaudiTrainingArguments(
            output_dir=output_dir,
            num_train_epochs=num_epochs,
            per_device_train_batch_size=per_device_train_batch_size,
            bf16=True,
            save_strategy="epoch",
            report_to="none",
            use_habana=True,
            use_lazy_mode=True,
            # create checkpoint directories
            save_on_each_node=True,
            # gaudi_config_name=gaudi_config_name,
        )
        gaudi_config = GaudiConfig(
            use_fused_adam=True,
            use_fused_clip_norm=True,
            use_torch_autocast=True,
        )
        trainer = GaudiSFTTrainer(
            model=model,
            train_dataset=train_dataset,
            eval_dataset=test_dataset,
            peft_config=peft_config,
            formatting_func=formatting_prompts_func,
            max_seq_length=max_seq_length,
            tokenizer=tokenizer,
            args=training_arguments,
            gaudi_config=gaudi_config,
        )
        generate_kwargs = {}
    else:
        training_arguments = TrainingArguments(
            output_dir=output_dir,
            num_train_epochs=num_epochs,
            per_device_train_batch_size=per_device_train_batch_size,
            fp16=use_fp16,
            bf16=not use_fp16,
            # use_ipex=True, # TODO CPU test this possible optimization
            use_cpu=model.device.type == "cpu",
            save_strategy="epoch",
            report_to="none",
            # options to reduce GPU memory usage and improve performance
            # https://huggingface.co/docs/transformers/perf_train_gpu_one
            # https://stackoverflow.com/a/75793317
            # torch_compile=True,
            # fp16=False,  # fp16 increases memory consumption 1.5x
            # gradient_accumulation_steps=8,
            # gradient_checkpointing=True,
            # eval_accumulation_steps=1,
            # per_device_eval_batch_size=1,
        )

        # https://huggingface.co/docs/trl/main/en/sft_trainer#trl.SFTTrainer
        trainer = SFTTrainer(
            model=model,
            train_dataset=train_dataset,
            eval_dataset=test_dataset,
            peft_config=peft_config,
            formatting_func=formatting_prompts_func,
            data_collator=collator,
            max_seq_length=max_seq_length,
            tokenizer=tokenizer,
            args=training_arguments,
        )
        generate_kwargs = {}

    print("LINUX_TRAIN.PY: TRAINING")
    trainer.train()

    model.config.use_cache = True

    print("LINUX_TRAIN.PY: RUNNING INFERENCE ON THE OUTPUT MODEL")

    for i, (d, assistant_old) in enumerate(
        zip(test_dataset, assistant_old_lst, strict=False)
    ):
        output = model_generate(d["user"], **generate_kwargs)
        assistant_new = output.split(response_template.strip())[-1].strip()
        assistant_expected = d["assistant"]

        print(f"\n===\ntest {i}\n===\n")
        print("\n===\nuser\n===\n")
        print(d["user"])
        print("\n===\nassistant_old\n===\n")
        print(assistant_old)
        print("\n===\nassistant_new\n===\n")
        print(assistant_new)
        print("\n===\nassistant_expected\n===\n")
        print(assistant_expected)

    print("LINUX_TRAIN.PY: MERGING ADAPTERS")
    model = trainer.model.merge_and_unload()
    model.save_pretrained(output_dir / "merged_model")

    print("LINUX_TRAIN.PY: FINISHED")
    return output_dir



# SPDX-License-Identifier: Apache-2.0
# pylint: disable=duplicate-code

# Standard
from pathlib import Path
import re

# Third Party
from click.testing import CliRunner
import pytest

# First Party
from instructlab import lab

TAXONOMY_BASE = "main"

TEST_CUSTOM_YAML_RULES = b"""empty-lines: relaxed

rules:
  line-length:
    max: 180
"""


class TestLabDiff:
    """Test collection for `ilab taxonomy diff` command."""

    @pytest.fixture(autouse=True)
    def _init_taxonomy(self, taxonomy_dir):
        self.taxonomy = taxonomy_dir

    def test_diff(self, cli_runner: CliRunner):
        untracked_file = "compositional_skills/new/qna.yaml"
        tracked_file = "compositional_skills/tracked/qna.yaml"
        self.taxonomy.create_untracked(untracked_file)
        self.taxonomy.add_tracked(tracked_file)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert untracked_file in result.output
        assert tracked_file not in result.output
        assert result.exit_code == 0

    def test_diff_empty_base(self, cli_runner: CliRunner):
        taxonomy_base = "empty"
        untracked_file = "compositional_skills/new/qna.yaml"
        tracked_file = "compositional_skills/tracked/qna.yaml"
        self.taxonomy.create_untracked(untracked_file)
        self.taxonomy.add_tracked(tracked_file)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                taxonomy_base,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert untracked_file in result.output
        assert tracked_file in result.output
        assert result.exit_code == 0

    def test_diff_rm_tracked(self, cli_runner: CliRunner):
        tracked_file = "compositional_skills/tracked/qna.yaml"
        self.taxonomy.add_tracked(tracked_file)
        self.taxonomy.remove_file(tracked_file)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert tracked_file not in result.output
        assert result.exit_code == 0

    def test_diff_invalid_ext(self, cli_runner: CliRunner):
        untracked_file = "compositional_skills/writing/new/qna.txt"
        self.taxonomy.create_untracked(untracked_file)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )

        assert untracked_file in self.taxonomy.untracked_files
        # Invalid extension is silently filtered out
        assert untracked_file not in result.output
        assert result.exit_code == 0

    def test_diff_yml_ext(self, cli_runner: CliRunner):
        untracked_file = "compositional_skills/writing/new/qna.yml"
        self.taxonomy.create_untracked(untracked_file)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )

        assert untracked_file in self.taxonomy.untracked_files
        # Invalid extension is filtered out but there is a warning
        assert re.search(
            f"{re.escape(untracked_file)}.*must be named 'qna\\.yaml'",
            result.output,
            re.MULTILINE,
        )
        assert not re.search(
            f"^{re.escape(untracked_file)}", result.output, re.MULTILINE
        )
        assert result.exit_code == 0

    def test_diff_YAML_ext(self, cli_runner: CliRunner):
        untracked_file = "compositional_skills/writing/new/qna.YAML"
        self.taxonomy.create_untracked(untracked_file)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )

        assert untracked_file in self.taxonomy.untracked_files
        # Invalid extension is filtered out but there is a warning
        assert re.search(
            f"{re.escape(untracked_file)}.*must be named 'qna\\.yaml'",
            result.output,
            re.MULTILINE,
        )
        assert not re.search(
            f"^{re.escape(untracked_file)}", result.output, re.MULTILINE
        )
        assert result.exit_code == 0

    def test_diff_invalid_base(self, cli_runner: CliRunner):
        taxonomy_base = "invalid"
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                taxonomy_base,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert (
            f'Couldn\'t find the taxonomy git ref "{taxonomy_base}" '
            "from the current HEAD" in result.output
        )
        assert result.exit_code == 1

    def test_diff_invalid_path(self, cli_runner: CliRunner):
        taxonomy_path = "/path/to/taxonomy"
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                taxonomy_path,
            ],
        )
        assert f"{taxonomy_path}" in result.output
        assert result.exit_code == 1

    def test_diff_valid_yaml(self, cli_runner: CliRunner, testdata_path: Path):
        qna = testdata_path.joinpath("skill_valid_answer.yaml").read_bytes()
        valid_yaml_file = "compositional_skills/valid/qna.yaml"
        self.taxonomy.create_untracked(valid_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert f"Taxonomy in {self.taxonomy.root} is valid :)" in result.output
        assert result.exit_code == 0

    def test_diff_valid_yaml_file(self, cli_runner: CliRunner, testdata_path):
        qna = testdata_path.joinpath("skill_valid_answer.yaml").read_bytes()
        valid_yaml_file = "compositional_skills/valid/qna.yaml"
        file_path = self.taxonomy.create_untracked(valid_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                file_path,
            ],
        )
        assert f"Taxonomy in {file_path} is valid :)" in result.output
        assert result.exit_code == 0

    def test_diff_valid_yaml_quiet(self, cli_runner: CliRunner, testdata_path):
        qna = testdata_path.joinpath("skill_valid_answer.yaml").read_bytes()
        valid_yaml_file = "compositional_skills/valid/qna.yaml"
        self.taxonomy.create_untracked(valid_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
                "--quiet",
            ],
        )
        assert result.output == ""
        assert result.exit_code == 0

    def test_diff_valid_yaml_quiet_file(self, cli_runner: CliRunner, testdata_path):
        qna = testdata_path.joinpath("skill_valid_answer.yaml").read_bytes()
        valid_yaml_file = "compositional_skills/valid/qna.yaml"
        file_path = self.taxonomy.create_untracked(valid_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                file_path,
                "--quiet",
            ],
        )
        assert result.output == ""
        assert result.exit_code == 0

    def test_diff_invalid_yaml_content(self, cli_runner: CliRunner, testdata_path):
        qna = testdata_path.joinpath("invalid_yaml.yaml").read_bytes()
        self.taxonomy.create_untracked("compositional_skills/invalid/qna.yaml", qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert "Reading taxonomy failed" in result.output
        assert result.exit_code == 1

    def test_diff_invalid_yaml_content_quiet(
        self, cli_runner: CliRunner, testdata_path
    ):
        qna = testdata_path.joinpath("invalid_yaml.yaml").read_bytes()
        self.taxonomy.create_untracked("compositional_skills/invalid/qna.yaml", qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
                "--quiet",
            ],
        )
        assert result.exception is not None
        assert result.exit_code == 1

    def test_diff_custom_yaml(self, cli_runner: CliRunner, testdata_path: Path):
        qna = testdata_path.joinpath("invalid_yaml.yaml").read_bytes()
        custom_rules_file = Path("custom_rules.yaml")
        custom_rules_file.write_bytes(TEST_CUSTOM_YAML_RULES)
        invalid_yaml_file = "compositional_skills/invalid/qna.yaml"
        self.taxonomy.create_untracked(invalid_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
                "--yaml-rules",
                str(custom_rules_file),
                "--quiet",
            ],
        )
        # custom yaml rules mean "invalid" yaml file should pass
        assert result.output == ""
        assert result.exit_code == 0
        Path.unlink(custom_rules_file)

    def test_diff_failing_schema_yaml(self, cli_runner: CliRunner, testdata_path: Path):
        qna = testdata_path.joinpath("skill_incomplete.yaml").read_bytes()
        failing_yaml_file = "compositional_skills/failing/qna.yaml"
        self.taxonomy.create_untracked(failing_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert "Reading taxonomy failed" in result.output
        assert result.exit_code == 1

    def test_diff_unsupported_knowledge_yaml(
        self, cli_runner: CliRunner, testdata_path: Path
    ):
        qna = testdata_path.joinpath("knowledge_unsupported.yaml").read_bytes()
        failing_yaml_file = "knowledge/unsupported/qna.yaml"
        self.taxonomy.create_untracked(failing_yaml_file, qna)
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert "Reading taxonomy failed" in result.output
        assert result.exit_code == 1

    def test_diff_invalid_yaml_filename(self, cli_runner: CliRunner, testdata_path):
        qna = testdata_path.joinpath("invalid_yaml.yaml").read_bytes()
        self.taxonomy.create_untracked(
            "compositional_skills/invalid/qna_invalid.yaml", qna
        )
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
            ],
        )
        assert f"Taxonomy in {self.taxonomy.root} is valid :)" in result.output
        assert result.exit_code == 0

    def test_diff_invalid_yaml_filename_quiet(
        self, cli_runner: CliRunner, testdata_path
    ):
        qna = testdata_path.joinpath("invalid_yaml.yaml").read_bytes()
        self.taxonomy.create_untracked(
            "compositional_skills/invalid/qna_invalid.yaml", qna
        )
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "taxonomy",
                "diff",
                "--taxonomy-base",
                TAXONOMY_BASE,
                "--taxonomy-path",
                self.taxonomy.root,
                "--quiet",
            ],
        )
        assert result.output == ""
        assert result.exit_code == 0



# SPDX-License-Identifier: MIT

# Standard
import os

# Third Party
from transformers import AutoModelForCausalLM, AutoTokenizer
import fire


def convert_bin_to_safetensors(input_dir, output_dir):
    input_dir, output_dir = (
        os.path.expanduser(input_dir),
        os.path.expanduser(output_dir),
    )

    model_path = os.path.expanduser(input_dir)

    tokenizer = AutoTokenizer.from_pretrained(model_path)
    model = AutoModelForCausalLM.from_pretrained(model_path)

    tokenizer.save_pretrained(output_dir)
    tokenizer.save_vocabulary(output_dir)
    model.save_pretrained(output_dir)

    # save_dict = model.state_dict()
    # torch.save(save_dict, os.path.join(output_dir, "pytorch_model.bin"))
    # model.config.to_json_file(os.path.join(output_dir, "config.json"))


if __name__ == "__main__":
    fire.Fire(convert_bin_to_safetensors)



# SPDX-License-Identifier: Apache-2.0

# Standard
import logging

# Third Party
import click

# First Party
from instructlab import clickext

logger = logging.getLogger(__name__)


@click.group(
    cls=clickext.LazyEntryPointGroup,
    ep_group="instructlab.command.system",
)
@click.pass_context
def system(ctx):
    """Command group for all system-related command calls"""
    ctx.obj = ctx.parent.obj
    ctx.default_map = ctx.parent.default_map



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
import enum
import logging
import os
import pathlib

# Third Party
# pylint: disable=ungrouped-imports
import click

# First Party
from instructlab import clickext
from instructlab.configuration import DEFAULTS, map_train_to_library

logger = logging.getLogger(__name__)

ADDITIONAL_ARGUMENTS = "additional_args"


class SupportedTrainingStrategies(enum.Enum):
    """Available advanced training strategies"""

    LAB_MULTIPHASE: str = "lab-multiphase"


def clickpath_setup(is_dir: bool) -> click.Path:
    """
    Creates a click.Path object meeting requirements:
        - path to target must exist,
        - file vs. directory are mutually exclusive,
        - path is fully resolved.
    """

    return click.Path(
        exists=True,
        file_okay=not is_dir,
        dir_okay=is_dir,
        resolve_path=True,
        path_type=pathlib.Path,
    )


@click.command()
@click.option(
    "--data-path",
    type=click.Path(file_okay=True),
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--ckpt-output-dir",
    type=click.Path(),
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--data-output-dir",
    type=click.Path(),
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--input-dir",
    type=click.Path(),
    show_default=True,  # TODO: set to None and change help message
    help="Path to generated files to use as input.",
)
@click.option(
    "--gguf-model-path",
    help="Local directory where gguf model is stored.",
    default=None,
    show_default=True,
)
@click.option(
    "--skip-preprocessing",
    help="Skips data preprocessing step for MLX training if data is already cached.",
    is_flag=True,
)
@click.option(
    "--tokenizer-dir",
    type=click.Path(),
    help="Base directory where tokenizer is stored.",
    default=None,
    show_default=True,
)
@click.option(
    "--model-path",
    type=click.Path(),
    cls=clickext.ConfigOption,
    required=True,  # default from config
    default=DEFAULTS.MODEL_REPO,
)
@click.option(
    "--iters",
    help="Number of iterations to train LoRA.",
    default=100,
)
@click.option(
    "--local",
    is_flag=True,
    help="Whether or not `model_dir` is remote from HuggingFace.",
)
@click.option(
    "-sq",
    "--skip-quantize",
    is_flag=True,
    help="Whether to skip quantization while converting to MLX. This parameter will be ignored if --gguf-model-path and --tokenizer-dir are specified.",
)
@click.option(
    "--num-epochs",
    type=click.INT,
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--device",
    type=click.Choice(["cpu", "cuda", "hpu", "mps"]),
    cls=clickext.ConfigOption,
)
@click.option(
    "--4-bit-quant",
    "four_bit_quant",
    is_flag=True,
    show_default=True,
    default=False,
    # TODO: hidden option until llamacpp_convert_to_gguf.py supports
    # quantized models, https://github.com/instructlab/instructlab/issues/579
    hidden=True,
    help=(
        "Use BitsAndBytes for 4-bit quantization "
        "(reduces GPU VRAM usage and may slow down training)"
    ),
)
@click.option(
    "--max-seq-len",
    type=int,
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--max-batch-len",
    type=int,
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--effective-batch-size",
    type=int,
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--save-samples",
    type=int,
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--learning-rate",
    type=float,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    required=True,  # default from config
)
@click.option(
    "--warmup-steps",
    type=int,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    required=True,  # default from config
)
@click.option(
    "--deepspeed-cpu-offload-optimizer",
    type=bool,
    cls=clickext.ConfigOption,
    required=True,  # default from config
    # config_section="deepspeed_options",
)
@click.option(
    "--deepspeed-cpu-offload-optimizer-ratio",
    type=float,
    cls=clickext.ConfigOption,
    required=True,  # default from config
    config_sections=ADDITIONAL_ARGUMENTS,
)
@click.option(
    "--deepspeed-cpu-offload-optimizer-pin-memory",
    type=bool,
    cls=clickext.ConfigOption,
    required=True,  # default from config
    config_sections=ADDITIONAL_ARGUMENTS,
)
@click.option(
    "--fsdp-cpu-offload-optimizer",
    type=bool,
    cls=clickext.ConfigOption,
)
@click.option(
    "--distributed-backend",
    type=str,
    cls=clickext.ConfigOption,
)
# below flags are invalid if lora == false
@click.option(
    "--lora-rank",
    type=int,
    cls=clickext.ConfigOption,
    # config_section="lora",
)
@click.option(
    "--lora-alpha",
    type=int,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
)
@click.option(
    "--lora-dropout",
    type=float,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
)
@click.option(
    "--lora-target-modules",
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    multiple=True,
    default=[],
)
@click.option(
    "--lora-quantize-dtype",
    type=str,
    cls=clickext.ConfigOption,
    default=None,
)
@click.option(
    "--is-padding-free",
    cls=clickext.ConfigOption,
    type=bool,
)
@click.option(
    "--gpus",
    "nproc_per_node",
    cls=clickext.ConfigOption,
    type=int,
)
@click.option(
    "--nnodes",
    type=int,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    required=True,  # default from config
)
@click.option(
    "--node-rank",
    type=int,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    required=True,  # default from config
)
@click.option(
    "--rdzv-id",
    type=int,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    required=True,  # default from config
)
@click.option(
    "--rdzv-endpoint",
    type=str,
    cls=clickext.ConfigOption,
    config_sections=ADDITIONAL_ARGUMENTS,
    required=True,  # default from config
)
@click.option("--disable-flash-attn", is_flag=True, cls=clickext.ConfigOption)
@click.option(
    "--strategy",
    type=click.Choice(
        [SupportedTrainingStrategies.LAB_MULTIPHASE.value], case_sensitive=False
    ),
    show_default=True,
    help="If chosen, will run the selected training strategy instead of a single training run.",
)
@click.option(
    "--phased-base-dir",
    type=clickpath_setup(is_dir=True),
    cls=clickext.ConfigOption,
)
@click.option(
    "--phased-phase1-data",
    type=clickpath_setup(is_dir=False),
    help="Path to .jsonl file that will be used for the first phase of end-to-end training.",
)
@click.option(
    "--phased-phase1-num-epochs",
    cls=clickext.ConfigOption,
    type=click.IntRange(min=1),
)
@click.option(
    "--phased-phase1-samples-per-save",
    cls=clickext.ConfigOption,
    type=click.IntRange(min=0),
)
@click.option(
    "--phased-phase1-effective-batch-size",
    cls=clickext.ConfigOption,
    type=click.IntRange(min=1),
)
@click.option(
    "--phased-phase2-data",
    type=clickpath_setup(is_dir=False),
    help="Path to .jsonl file that will be used for the second phase of end-to-end training.",
)
@click.option(
    "--phased-phase2-num-epochs",
    cls=clickext.ConfigOption,
    type=click.IntRange(min=1),
)
@click.option(
    "--phased-phase2-samples-per-save",
    cls=clickext.ConfigOption,
    type=click.IntRange(min=0),
)
@click.option(
    "--phased-phase2-effective-batch-size",
    cls=clickext.ConfigOption,
    type=click.IntRange(min=1),
)
@click.option(
    "--phased-mt-bench-judge",
    # type=clickpath_setup(is_dir=True), # want this in the future, can't guarantee it exists so can't enforce it this way.
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
    cls=clickext.ConfigOption,
)
@click.option(
    "--skip-user-confirm",
    "-y",
    is_flag=True,
    help="Skips any user confirmation prompts.",
)
@click.option(
    "--enable-serving-output",
    is_flag=True,
    help="Print serving engine logs during phased training checkpoint evaluation.",
)
@click.option(
    "--checkpoint-at-epoch",
    is_flag=True,
    help="By default, checkpoints are saved at the end of each training epoch. This option disables this behavior.",
)
@click.option(
    "--pipeline",
    type=click.Choice(["simple", "full", "accelerated"]),
    cls=clickext.ConfigOption,
)
@click.option(
    "--training-journal",
    cls=clickext.ConfigOption,
)
@click.option(
    "--force-clear-phased-cache",
    is_flag=True,
    help="Clear phased cache (journal, checkpoints, metadata). Helpful paired with '--skip-user-confirm'",
)
@click.pass_context
@clickext.display_params
def train(
    ctx,
    data_path: str,
    input_dir,
    skip_preprocessing,
    tokenizer_dir,
    gguf_model_path,
    model_path,
    iters,
    local,
    skip_quantize,
    num_epochs,
    device: str,
    four_bit_quant: bool,
    strategy: str | None,
    phased_base_dir: pathlib.Path,
    phased_phase1_data: pathlib.Path | None,
    phased_phase1_num_epochs: int | None,
    phased_phase1_samples_per_save: int | None,
    phased_phase1_effective_batch_size: int | None,
    phased_phase2_data: pathlib.Path | None,
    phased_phase2_num_epochs: int | None,
    phased_phase2_samples_per_save: int | None,
    phased_phase2_effective_batch_size: int | None,
    phased_mt_bench_judge: pathlib.Path | None,
    skip_user_confirm: bool,
    enable_serving_output: bool,
    pipeline: str,
    training_journal: pathlib.Path | None,
    force_clear_phased_cache: bool,
    distributed_backend,
    **kwargs,
):
    """
    Takes synthetic data generated locally with `ilab data generate` and the previous model and learns a new model using the MLX API.
    On success, writes newly learned model to {model_dir}/mlx_model, which is where `chatmlx` will look for a model.
    """
    if (
        pipeline in ("full", "simple")
        and strategy == SupportedTrainingStrategies.LAB_MULTIPHASE.value
    ):
        ctx.fail("Multi Phase training is only supported with `--pipeline accelerated`")

    # TODO: cdoern, remove this flag
    if not input_dir:
        # By default, generate output-dir is used as train input-dir
        input_dir = ctx.obj.config.generate.output_dir

    if four_bit_quant and device != "cuda":
        ctx.fail("'--4-bit-quant' option requires '--device=cuda'")

    if (
        pipeline in ("full", "accelerated")
    ) and strategy != SupportedTrainingStrategies.LAB_MULTIPHASE.value:
        if not os.path.isfile(data_path):
            ctx.fail(
                f"Data path must be to a valid .jsonl file. Value given: {data_path}"
            )
    # we can use train_args locally to run lower fidelity training
    if is_high_fidelity(device=device) and pipeline == "accelerated":
        train_args, torch_args = map_train_to_library(ctx, ctx.params)

        # First Party
        from instructlab.model import accelerated_train

        try:
            accelerated_train.accelerated_train(
                train_args=train_args,
                torch_args=torch_args,
                strategy=strategy,
                distributed_backend=distributed_backend,
                phased_phase1_data=phased_phase1_data,
                phased_phase2_data=phased_phase2_data,
                phased_base_dir=phased_base_dir,
                phased_phase1_num_epochs=phased_phase1_num_epochs,
                phased_phase1_samples_per_save=phased_phase1_samples_per_save,
                phased_phase1_effective_batch_size=phased_phase1_effective_batch_size,
                phased_phase2_num_epochs=phased_phase2_num_epochs,
                phased_phase2_samples_per_save=phased_phase2_samples_per_save,
                phased_phase2_effective_batch_size=phased_phase2_effective_batch_size,
                enable_serving_output=enable_serving_output,
                phased_mt_bench_judge=phased_mt_bench_judge,
                skip_user_confirm=skip_user_confirm,
                force_clear_phased_cache=force_clear_phased_cache,
                eval_serve=ctx.obj.config.serve,
                eval_gpus=ctx.obj.config.evaluate.gpus,
                training_journal=training_journal,
            )
        except Exception as exc:
            click.secho(f"Accelerated Training failed with {str(exc)}")
            raise click.exceptions.Exit(1)
    elif not is_high_fidelity(device=device) and pipeline == "full":
        # Third Party
        import torch

        torch.set_autocast_enabled(False)
        # First Party
        from instructlab.model import full_train

        train_args, torch_args = map_train_to_library(ctx, ctx.params)
        # if on CPU or MPS, execute full train, which is based
        # off of the structure of the training repo, just with different optimizers, model sizes, and special data gradient accumulation to get it
        # to fit on most consumer laptops
        full_train.train(train_args=train_args, device=device)
    elif pipeline == "simple":
        # First Party
        from instructlab.model import simple_train

        try:
            simple_train.simple_train(
                model_path=model_path,
                skip_preprocessing=skip_preprocessing,
                skip_quantize=skip_quantize,
                gguf_model_path=gguf_model_path,
                tokenizer_dir=tokenizer_dir,
                data_path=data_path,
                input_dir=input_dir,
                ckpt_output_dir=Path(kwargs["ckpt_output_dir"]),
                iters=iters,
                local=local,
                num_epochs=num_epochs,
                device=device,
                four_bit_quant=four_bit_quant,
            )
        except Exception as exc:
            click.secho(f"{exc}", fg="red")
            raise click.exceptions.Exit(1)


# chooses which type of training to run depending on the device provided
def is_high_fidelity(device):
    return device in ("cuda", "hpu")



# SPDX-License-Identifier: Apache-2.0

# pylint: disable=too-many-lines

# Standard
import multiprocessing

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab import configuration as cfg

# 'fork' is unsafe and incompatible with some hardware accelerators.
# Python 3.14 will switch to 'spawn' on all platforms.
multiprocessing.set_start_method(cfg.DEFAULTS.MULTIPROCESSING_START_METHOD, force=True)


@click.group(
    cls=clickext.ExpandAliasesGroup,
    ep_group="instructlab.command",
    alias_ep_group="instructlab.command.alias",
)
@click.option(
    "--config",
    "config_file",
    type=click.Path(),
    default=cfg.DEFAULTS.CONFIG_FILE,
    show_default=True,
    help="Path to a configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    "debug_level",
    count=True,
    default=0,
    show_default=False,
    help="Enable debug logging (repeat for even more verbosity)",
)
@click.version_option(package_name="instructlab")
@click.pass_context
# pylint: disable=redefined-outer-name
def ilab(ctx, config_file, debug_level: int = 0):
    """CLI for interacting with InstructLab.

    If this is your first time running ilab, it's best to start with `ilab config init` to create the environment.
    """
    cfg.init(ctx, config_file, debug_level)



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
from time import time
from typing import Any, Dict
import logging
import os

# Third Party
from datasets import load_dataset
from openai import OpenAI, Stream
import click

# First Party
from instructlab.configuration import DEFAULTS

# Local
from ..utils import get_sysprompt, http_client
from .backends import backends

logger = logging.getLogger(__name__)


def response(client, user: str, create_params: dict):
    # https://platform.openai.com/docs/api-reference/chat/create
    logger.debug("sending %d bytes", len(user))
    start_time = time()
    resp = client.chat.completions.create(
        **create_params,
        messages=[
            {"role": "system", "content": get_sysprompt()},
            {"role": "user", "content": user},
        ],
    )
    if isinstance(resp, Stream):
        # https://platform.openai.com/docs/api-reference/chat/streaming
        resp = ""
        for r in resp:
            if c := r.choices[0].delta.content:
                print(c, end="", flush=True)
                resp += c
        print()
        return resp
    # https://platform.openai.com/docs/api-reference/chat/object
    c = resp.choices[0].message.content
    logger.debug("received %d bytes, elapsed %.1fs", len(c), time() - start_time)
    return c


def test_model(ctx, res, ds, model: Path, create_params: dict):
    logger.debug("%s", model)
    create_params["model"] = str(model)  # mandatory, but not used
    # TODO: this code block is replicated again. Refactor to a common function.
    backend_instance = None
    try:
        ctx.obj.config.serve.llama_cpp.llm_family = ctx.params["model_family"]
        backend_instance = backends.select_backend(
            cfg=ctx.obj.config.serve, model_path=model
        )
        try:
            api_base = backend_instance.run_detached(http_client(ctx.params))
        except Exception as exc:
            click.secho(f"Failed to start server: {exc}", fg="red")
            raise click.exceptions.Exit(1)
        api_base = api_base or ctx.obj.config.serve.api_base()
        logger.debug("api_base=%s", api_base)
        client = OpenAI(
            base_url=api_base,
            api_key=ctx.params["api_key"],
        )
        for d in ds:
            res[d["user"]][str(model)] = response(client, d["user"], create_params)
    except Exception as exc:
        raise exc
    finally:
        if backend_instance:
            backend_instance.shutdown()


def linux_test(
    ctx: click.Context,
    test_file: Path,
    models=None,
    create_params=None,
) -> Dict[str, Any]:
    # linux_test
    logger.debug("test_file=%s", test_file)
    if not models:
        models = [DEFAULTS.DEFAULT_CHAT_MODEL]
    if not create_params:
        create_params = {}

    # https://platform.openai.com/docs/api-reference/chat/create
    create_params["temperature"] = 0  # for more reproducible results
    # create_params["stream"] = True  # optional, useful for interactive debugging

    ds = load_dataset("json", data_files=os.fspath(test_file), split="train")
    res: dict[str, Any] = {}
    # intentionally not using collections.defaultdict to avoid dependency
    for d in ds:
        res[d["user"]] = {}
    for m in models:
        test_model(ctx, res, ds, m, create_params)
    return res



# SPDX-License-Identifier: MIT
# Copyright © 2023 Apple Inc.

# Standard
from pathlib import Path
from typing import Any, Generator
import glob
import json
import logging
import os

# Third Party
from huggingface_hub import snapshot_download
from safetensors.torch import save_file
import mlx.core as mx
import mlx.nn as nn
import transformers

# Local
from .models import llama, mixtral, phi2

# Constants
MODEL_MAPPING = {
    "llama": llama,
    "mistral": llama,  # mistral is compatible with llama
    "phi": phi2,
    "mixtral": mixtral,
}


def _get_classes(config: dict):
    """
    Retrieve the model and model args classes based on the configuration.

    Args:
        config (dict): The model configuration.

    Returns:
        A tuple containing the Model class and the ModelArgs class.
    """
    model_type = config["model_type"]
    if model_type not in MODEL_MAPPING:
        msg = f"Model type {model_type} not supported."
        logging.error(msg)
        raise ValueError(msg)
    print(f"Using {model_type=}")
    arch = MODEL_MAPPING[model_type]
    return arch.Model, arch.ModelArgs


def fetch_from_hub(hf_path: str, local: bool):
    if local:
        model_path = hf_path
    else:
        model_path = snapshot_download(
            repo_id=hf_path,
            local_dir=hf_path.replace(
                "/", "-"
            ),  # "instructlab/merlinite-7b-lab" to "instructlab-merlinite-7b-lab"
            allow_patterns=["*.json", "*.safetensors", "tokenizer.model"],
        )

    weight_files = glob.glob(f"{model_path}/*.safetensors")
    if len(weight_files) == 0:
        raise FileNotFoundError("No safetensors found in {}".format(model_path))

    weights = {}
    for wf in weight_files:
        w = mx.load(wf, return_metadata=False)
        weights.update(w.items())

    config = transformers.AutoConfig.from_pretrained(hf_path)
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        hf_path,
    )
    return weights, config.to_dict(), tokenizer


def upload_to_hub(path: str, name: str, hf_path: str):
    # Standard
    import os

    # Third Party
    from huggingface_hub import HfApi, ModelCard, logging

    repo_id = f"mlx-community/{name}"

    card = ModelCard.load(hf_path)
    card.data.tags = ["mlx"] if card.data.tags is None else card.data.tags + ["mlx"]
    card.text = f"""
# {name}
This model was converted to MLX format from [`{hf_path}`]().
Refer to the [original model card](https://huggingface.co/{hf_path}) for more details on the model.
## Use with mlx
```bash
pip install mlx
git clone https://github.com/ml-explore/mlx-examples.git
cd mlx-examples/llms/hf_llm
python generate.py --model {repo_id} --prompt "My name is"
```
"""
    card.save(os.path.join(path, "README.md"))

    logging.set_verbosity_info()

    api = HfApi()
    api.create_repo(repo_id=repo_id, exist_ok=True)
    api.upload_folder(
        folder_path=path,
        repo_id=repo_id,
        repo_type="model",
    )


def make_shards(weights: dict, max_file_size_gibibyte: int = 15):
    max_file_size_bytes = max_file_size_gibibyte << 30
    shards = []
    shard: dict[str, Any] = {}
    shard_size = 0
    for k, v in weights.items():
        if shard_size + v.nbytes > max_file_size_bytes:
            shards.append(shard)
            shard, shard_size = {}, 0
        shard[k] = v
        shard_size += v.nbytes
    shards.append(shard)
    return shards


def save_model(save_dir: str, weights, tokenizer, config):
    save_dir_path = Path(save_dir)
    save_dir_path.mkdir(parents=True, exist_ok=True)

    dtype = weights[next(iter(weights.keys()))].dtype
    if str(dtype) in [str(dtype) for dtype in [mx.float16, mx.bfloat16, mx.float32]]:
        shards = make_shards(weights, max_file_size_gibibyte=5)
        shards_count = len(shards)
        shard_file_format = (
            "model-{:05d}-of-{:05d}.safetensors"
            if shards_count > 1
            else "model.safetensors"
        )

        for i, shard in enumerate(shards):
            shard_name = shard_file_format.format(i + 1, shards_count)
            mx.save_safetensors(str(save_dir_path / shard_name), shard)
    else:
        save_file(
            weights,
            os.path.join(save_dir_path, "model.safetensors"),
            metadata={"format": "pt"},
        )

    tokenizer.save_pretrained(save_dir_path)
    tokenizer.save_vocabulary(save_dir_path)

    with open(save_dir_path / "config.json", "w") as fid:
        json.dump(config, fid, indent=4)


def load(path_or_hf_repo: str):
    # If the path exists, it will try to load model form it
    # otherwise download and cache from the hf_repo and cache
    model_path = Path(os.path.expandvars(os.path.expanduser(path_or_hf_repo)))
    if not model_path.exists():
        model_path = Path(
            snapshot_download(
                repo_id=path_or_hf_repo,
                allow_patterns=["*.json", "*.safetensors", "tokenizer.model"],
            )
        )

    with open(model_path / "config.json", "r") as f:
        config = json.loads(f.read())
        quantization = config.get("quantization", None)

    weight_files = glob.glob(str(model_path / "*.safetensors"))
    if len(weight_files) == 0:
        raise FileNotFoundError("No safetensors found in {}".format(model_path))

    weights = {}
    for wf in weight_files:
        w = mx.load(wf, return_metadata=False)
        weights.update(w.items())

    model_class, model_args_class = _get_classes(config=config)
    model_args = model_args_class.from_dict(config)
    model = model_class(model_args)
    if quantization is not None:
        nn.QuantizedLinear.quantize_module(
            model,
            **quantization,
            linear_class_predicate=lambda m: isinstance(m, nn.Linear)
            and m.weight.shape[0] != 8,
        )

    model.load_weights(list(weights.items()))

    mx.eval(model.parameters())
    tokenizer = transformers.AutoTokenizer.from_pretrained(model_path)
    return model, tokenizer, config


def generate(
    prompt: mx.array, model: nn.Module, temp: float = 0.0
) -> Generator[mx.array, None, None]:
    """
    Generate text based on the given prompt and model.

    Args:
        prompt (mx.array): The input prompt.
        model (nn.Module): The model to use for generation.
        temp (float): The temperature for sampling. If temp is 0, use max sampling.

    Yields:
        mx.array: The generated text.
    """

    def sample(logits: mx.array) -> mx.array:
        return (
            mx.argmax(logits, axis=-1)
            if temp == 0
            else mx.random.categorical(logits * (1 / temp))
        )

    y = prompt
    cache = None
    while True:
        logits, cache = model(y[None], cache=cache)
        logits = logits[:, -1, :]
        y = sample(logits)
        yield y



# Standard

# Third Party
from click.testing import CliRunner

# Local
from . import common


def test_vllm_args_null(cli_runner: CliRunner):
    fname = common.setup_gpus_config(
        section_path="generate.teacher", vllm_args=lambda: None
    )
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "data",
            "generate",
            "--gpus",
            "4",
        ],
    )
    common.assert_tps(args, "4")



# SPDX-License-Identifier: MIT

# Standard
from pathlib import Path
import os
import sys

# Third Party
import click

# Get the absolute path of the current script
current_script_path = os.path.abspath(__file__)
# Get the parent directory of the current script
parent_directory = os.path.dirname(current_script_path)
# Add the parent directory to sys.path
sys.path.append(parent_directory)

# Third Party
from huggingface_hub import snapshot_download
from safetensors.torch import save_file
import mlx.core as mx
import transformers

# First Party
from instructlab.utils import macos_requirement


def fetch_tokenizer_from_hub(hf_path: str, local_dir: str):
    model_path = snapshot_download(
        repo_id=hf_path,
        local_dir=local_dir,
        allow_patterns=["*.json", "tokenizer.model"],
    )

    config = transformers.AutoConfig.from_pretrained(hf_path)
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        hf_path,
    )
    return config.to_dict(), tokenizer


def make_shards(weights: dict, max_file_size_gibibyte: int = 15):
    max_file_size_bytes = max_file_size_gibibyte << 30
    shards = []
    shard, shard_size = {}, 0
    for k, v in weights.items():
        if shard_size + v.nbytes > max_file_size_bytes:
            shards.append(shard)
            shard, shard_size = {}, 0
        shard[k] = v
        shard_size += v.nbytes
    shards.append(shard)
    return shards


@macos_requirement(echo_func=click.secho, exit_exception=click.exceptions.Exit)
def save_model(save_dir: str, weights):
    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    dtype = weights[next(iter(weights.keys()))].dtype
    if str(dtype) in [str(dtype) for dtype in [mx.float16, mx.bfloat16, mx.float32]]:
        shards = make_shards(weights, max_file_size_gibibyte=5)
        shards_count = len(shards)
        shard_file_format = (
            "model-{:05d}-of-{:05d}.safetensors"
            if shards_count > 1
            else "model.safetensors"
        )

        for i, shard in enumerate(shards):
            shard_name = shard_file_format.format(i + 1, shards_count)
            mx.save_safetensors(str(save_dir / shard_name), shard)
    else:
        save_file(
            weights,
            os.path.join(save_dir, "model.safetensors"),
            metadata={"format": "pt"},
        )
